@@ -74,14 +74,20 @@ sources_selected (GDBusConnection *bus,
   guint32 response;
   g_autoptr(GVariant) ret = NULL;
 
+  if (call->cancelled_id)
+    {
+      g_signal_handler_disconnect (g_task_get_cancellable (call->task), call->cancelled_id);
+      call->cancelled_id = 0;
+    }
+
   g_variant_get (parameters, "(u@a{sv})", &response, &ret);
 
   if (response == 0)
     {
-       XdpSession *session;
+      XdpSession *session;
 
-       session = _xdp_session_new (call->portal, call->id, call->type);
-       g_task_return_pointer (call->task, session, g_object_unref);
+      session = _xdp_session_new (call->portal, call->id, call->type);
+      g_task_return_pointer (call->task, session, g_object_unref);
     }
   else if (response == 1)
     g_task_return_new_error (call->task, G_IO_ERROR, G_IO_ERROR_CANCELLED, "Screencast SelectSources() canceled");
@@ -103,6 +109,12 @@ call_returned (GObject *object,
   ret = g_dbus_connection_call_finish (G_DBUS_CONNECTION (object), result, &error);
   if (error)
     {
+      if (call->cancelled_id)
+        {
+          g_signal_handler_disconnect (g_task_get_cancellable (call->task), call->cancelled_id);
+          call->cancelled_id = 0;
+        }
+
       g_task_return_error (call->task, error);
       create_call_free (call);
     }
@@ -168,18 +180,30 @@ devices_selected (GDBusConnection *bus,
 
   g_variant_get (parameters, "(u@a{sv})", &response, &ret);
 
+  if (response != 0 && call->cancelled_id)
+    {
+      g_signal_handler_disconnect (g_task_get_cancellable (call->task), call->cancelled_id);
+      call->cancelled_id = 0;
+    }
+
   if (response == 0)
     {
       g_dbus_connection_signal_unsubscribe (call->portal->bus, call->signal_id);
       call->signal_id = 0;
 
-      if (call->outputs != 0)
+      if (call->outputs != XDP_OUTPUT_NONE)
         select_sources (call);
       else
-       {
-         g_task_return_pointer (call->task, _xdp_session_new (call->portal, call->id, call->type), g_object_unref);
-         create_call_free (call);
-       }
+        {
+          if (call->cancelled_id)
+            {
+              g_signal_handler_disconnect (g_task_get_cancellable (call->task), call->cancelled_id);
+              call->cancelled_id = 0;
+            }
+
+          g_task_return_pointer (call->task, _xdp_session_new (call->portal, call->id, call->type), g_object_unref);
+          create_call_free (call);
+        }
     }
   else if (response == 1)
     g_task_return_new_error (call->task, G_IO_ERROR, G_IO_ERROR_CANCELLED, "Remote desktop SelectDevices() canceled");
@@ -242,6 +266,12 @@ session_created (GDBusConnection *bus,
 
   g_variant_get (parameters, "(u@a{sv})", &response, &ret);
 
+  if (response != 0 && call->cancelled_id)
+    {
+      g_signal_handler_disconnect (g_task_get_cancellable (call->task), call->cancelled_id);
+      call->cancelled_id = 0;
+    }
+
   if (response == 0)
     {
       g_dbus_connection_signal_unsubscribe (call->portal->bus, call->signal_id);
@@ -327,9 +357,50 @@ create_session (CreateCall *call)
 }
 
 static void
-get_version_returned (GObject *object,
-                      GAsyncResult *result,
-                      gpointer data)
+get_remote_desktop_interface_version_returned (GObject *object,
+                                               GAsyncResult *result,
+                                               gpointer data)
+{
+  CreateCall *call = data;
+  GError *error = NULL;
+  g_autoptr(GVariant) version_variant = NULL;
+  g_autoptr(GVariant) ret = NULL;
+
+  ret = g_dbus_connection_call_finish (G_DBUS_CONNECTION (object), result, &error);
+  if (error)
+    {
+      g_task_return_error (call->task, error);
+      create_call_free (call);
+      return;
+    }
+
+  g_variant_get_child (ret, 0, "v", &version_variant);
+  call->portal->remote_desktop_interface_version = g_variant_get_uint32 (version_variant);
+
+  create_session (call);
+}
+
+static void
+get_remote_desktop_interface_version (CreateCall *call)
+{
+  g_dbus_connection_call (call->portal->bus,
+                          PORTAL_BUS_NAME,
+                          PORTAL_OBJECT_PATH,
+                          "org.freedesktop.DBus.Properties",
+                          "Get",
+                          g_variant_new ("(ss)", "org.freedesktop.portal.RemoteDesktop", "version"),
+                          NULL,
+                          G_DBUS_CALL_FLAGS_NONE,
+                          -1,
+                          g_task_get_cancellable (call->task),
+                          get_remote_desktop_interface_version_returned,
+                          call);
+}
+
+static void
+get_screencast_interface_version_returned (GObject *object,
+                                           GAsyncResult *result,
+                                           gpointer data)
 {
   CreateCall *call = data;
   GError *error = NULL;
@@ -347,7 +418,7 @@ get_version_returned (GObject *object,
   g_variant_get_child (ret, 0, "v", &version_variant);
   call->portal->screencast_interface_version = g_variant_get_uint32 (version_variant);
 
-  create_session (call);
+  get_remote_desktop_interface_version (call);
 }
 
 static void
@@ -363,10 +434,24 @@ get_screencast_interface_version (CreateCall *call)
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           g_task_get_cancellable (call->task),
-                          get_version_returned,
+                          get_screencast_interface_version_returned,
                           call);
 }
 
+static void
+get_portal_interface_versions (CreateCall *call)
+{
+  g_assert (call != NULL);
+  g_assert (call->portal != NULL);
+
+  if (call->portal->screencast_interface_version == 0)
+    get_screencast_interface_version (call);
+  else if (call->portal->remote_desktop_interface_version == 0)
+    get_remote_desktop_interface_version (call);
+  else
+    create_session (call);
+}
+
 /**
  * xdp_portal_create_screencast_session:
  * @portal: a [class@Portal]
@@ -411,10 +496,7 @@ xdp_portal_create_screencast_session (XdpPortal *portal,
   call->multiple = (flags & XDP_SCREENCAST_FLAG_MULTIPLE) != 0;
   call->task = g_task_new (portal, cancellable, callback, data);
 
-  if (portal->screencast_interface_version == 0)
-    get_screencast_interface_version (call);
-  else
-    create_session (call);
+  get_portal_interface_versions (call);
 }
 
 /**
@@ -423,7 +505,7 @@ xdp_portal_create_screencast_session (XdpPortal *portal,
  * @result: a [iface@Gio.AsyncResult]
  * @error: return location for an error
  *
- * Finishes the create-screencast request, and returns an [class@Session].
+ * Finishes the create-screencast request, and returns a [class@Session].
  *
  * Returns: (transfer full): a [class@Session]
  */
@@ -486,7 +568,7 @@ xdp_portal_create_remote_desktop_session (XdpPortal *portal,
   call->multiple = (flags & XDP_REMOTE_DESKTOP_FLAG_MULTIPLE) != 0;
   call->task = g_task_new (portal, cancellable, callback, data);
 
-  create_session (call);
+  get_portal_interface_versions (call);
 }
 
 /**
@@ -495,7 +577,7 @@ xdp_portal_create_remote_desktop_session (XdpPortal *portal,
  * @result: a [iface@Gio.AsyncResult]
  * @error: return location for an error
  *
- * Finishes the create-remote-desktop request, and returns an [class@Session].
+ * Finishes the create-remote-desktop request, and returns a [class@Session].
  *
  * Returns: (transfer full): a [class@Session]
  */
@@ -567,6 +649,12 @@ session_started (GDBusConnection *bus,
   guint32 response;
   g_autoptr(GVariant) ret = NULL;
 
+  if (call->cancelled_id)
+    {
+      g_signal_handler_disconnect (g_task_get_cancellable (call->task), call->cancelled_id);
+      call->cancelled_id = 0;
+    }
+
   g_variant_get (parameters, "(u@a{sv})", &response, &ret);
 
   _xdp_session_set_session_state (call->session, response == 0 ? XDP_SESSION_ACTIVE
@@ -766,9 +854,11 @@ xdp_session_close (XdpSession *session)
  * @session: a [class@Session]
  *
  * Opens a file descriptor to the pipewire remote where the screencast
- * streams are available. The file descriptor should be used to create
- * a pw_remote object, by using pw_remote_connect_fd(). Only the
- * screencast stream nodes will be available from this pipewire node.
+ * streams are available.
+ *
+ * The file descriptor should be used to create a pw_remote object, by using
+ * pw_remote_connect_fd(). Only the screencast stream nodes will be available
+ * from this pipewire node.
  *
  * Returns: the file descriptor
  */
@@ -808,6 +898,90 @@ xdp_session_open_pipewire_remote (XdpSession *session)
   return g_unix_fd_list_get (fd_list, fd_out, NULL);
 }
 
+static inline gboolean
+is_active_remote_desktop_session (XdpSession    *session,
+                                  XdpDeviceType  required_device)
+{
+  return XDP_IS_SESSION (session) &&
+         session->type == XDP_SESSION_REMOTE_DESKTOP &&
+         session->state == XDP_SESSION_ACTIVE &&
+         !session->uses_eis &&
+         (session->devices & required_device) != 0;
+}
+
+/**
+ * xdp_session_connect_to_eis
+ * @session: a [class@Session]
+ * @error: return location for a #GError pointer
+ *
+ * Connect this XdpRemoteDesktopSession to an EIS implementation and return the fd.
+ * This fd can be passed into ei_setup_backend_fd(). See the libei
+ * documentation for details.
+ *
+ * This call must be issued before xdp_session_start(). If successful, all input
+ * event emulation must be handled via the EIS connection and calls to
+ * xdp_session_pointer_motion() etc. are silently ignored.
+ *
+ * Returns: the file descriptor to the EIS implementation
+ */
+int
+xdp_session_connect_to_eis (XdpSession  *session,
+                            GError     **error)
+{
+  XdpPortal *portal = session->portal;
+  GVariantBuilder options;
+  g_autoptr(GVariant) ret = NULL;
+  g_autoptr(GUnixFDList) fd_list = NULL;
+  int fd_out = -1;
+
+  if (portal->remote_desktop_interface_version < 2)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED, "Not supported by the portal interface");
+      return -1;
+    }
+  else if (session->type != XDP_SESSION_REMOTE_DESKTOP)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT, "Session is not a Remote Desktop session");
+      return -1;
+    }
+  else if (xdp_session_get_session_state (session) != XDP_SESSION_ACTIVE)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT, "Session has not been started");
+      return -1;
+    }
+  else if (session->uses_eis)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT, "Session is already connected to EIS");
+      return -1;
+    }
+
+  g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
+
+  ret = g_dbus_connection_call_with_unix_fd_list_sync (portal->bus,
+                                                       PORTAL_BUS_NAME,
+                                                       PORTAL_OBJECT_PATH,
+                                                       "org.freedesktop.portal.RemoteDesktop",
+                                                       "ConnectToEIS",
+                                                       g_variant_new ("(oa{sv})",
+                                                                      session->id,
+                                                                      &options),
+                                                       NULL,
+                                                       G_DBUS_CALL_FLAGS_NONE,
+                                                       -1,
+                                                       NULL,
+                                                       &fd_list,
+                                                       NULL,
+                                                       error);
+  if (!ret)
+      return -1;
+
+  session->uses_eis = TRUE;
+
+  g_variant_get (ret, "(h)", &fd_out);
+
+  return g_unix_fd_list_get (fd_list, fd_out, NULL);
+}
+
 /**
  * xdp_session_pointer_motion:
  * @session: a [class@Session]
@@ -826,10 +1000,7 @@ xdp_session_pointer_motion (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_POINTER) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_POINTER));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -862,10 +1033,7 @@ xdp_session_pointer_position (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_POINTER) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_POINTER));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -895,10 +1063,7 @@ xdp_session_pointer_button (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_POINTER) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_POINTER));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -932,10 +1097,7 @@ xdp_session_pointer_axis (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_POINTER) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_POINTER));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_variant_builder_add (&options, "{sv}", "finish", g_variant_new_boolean (finish));
@@ -966,10 +1128,7 @@ xdp_session_pointer_axis_discrete (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_POINTER) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_POINTER));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -1001,10 +1160,7 @@ xdp_session_keyboard_key (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_KEYBOARD) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_KEYBOARD));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -1024,9 +1180,10 @@ xdp_session_keyboard_key (XdpSession *session,
  * @x: new X position
  * @y: new Y position
  *
- * Notify about a new touch down event. The `(x, y)` position
- * represents the new touch point position in the streams logical
- * coordinate space.
+ * Notify about a new touch down event.
+ *
+ * The `(x, y)` position represents the new touch point position in the streams
+ * logical coordinate space.
  *
  * May only be called on a remote desktop session
  * with `XDP_DEVICE_TOUCHSCREEN` access.
@@ -1040,10 +1197,7 @@ xdp_session_touch_down (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_TOUCHSCREEN) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_TOUCHSCREEN));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -1063,9 +1217,10 @@ xdp_session_touch_down (XdpSession *session,
  * @x: new X position
  * @y: new Y position
  *
- * Notify about a new touch motion event. The `(x, y)` position
- * represents where the touch point position in the streams logical
- * coordinate space moved.
+ * Notify about a new touch motion event.
+ *
+ * The `(x, y)` position represents where the touch point position in the
+ * streams logical coordinate space moved.
  *
  * May only be called on a remote desktop session
  * with `XDP_DEVICE_TOUCHSCREEN` access.
@@ -1079,10 +1234,7 @@ xdp_session_touch_position (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_TOUCHSCREEN) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_TOUCHSCREEN));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
@@ -1110,17 +1262,14 @@ xdp_session_touch_up (XdpSession *session,
 {
   GVariantBuilder options;
 
-  g_return_if_fail (XDP_IS_SESSION (session) &&
-                    session->type == XDP_SESSION_REMOTE_DESKTOP &&
-                    session->state == XDP_SESSION_ACTIVE &&
-                    ((session->devices & XDP_DEVICE_TOUCHSCREEN) != 0));
+  g_return_if_fail (is_active_remote_desktop_session (session, XDP_DEVICE_TOUCHSCREEN));
 
   g_variant_builder_init (&options, G_VARIANT_TYPE_VARDICT);
   g_dbus_connection_call (session->portal->bus,
                           PORTAL_BUS_NAME,
                           PORTAL_OBJECT_PATH,
                           "org.freedesktop.portal.RemoteDesktop",
-                          "NotifyTouchMotion",
+                          "NotifyTouchUp",
                           g_variant_new ("(oa{sv}u)", session->id, &options, slot),
                           NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
 }
